; 
; Author: Bradley A. Minch
; Organization: Franklin W. Olin College of Engineering
; Revision History: 
;     06/22/2005 - Added code to disable all endpoints on USRTIF and to mask
;                  bits 0, 1, and 7 of USTAT on TRNIF in ServiceUSB.
;     04/21/2005 - Initial public release.
;
; ============================================================================
; 
; Peripheral Description:
; 
; This peripheral enumerates as a USB keyboard (an HID class device), which 
; continually sends the keycodes for 'f', 'o', 'o', 'b', 'a', 'r', and ' ' over 
; and over again.  It also has a CAPS LOCK key (a switch on RA0) and status LED 
; (an LED on RA1).  The firmware is currently configured to use an external 
; 4-MHz crystal, to operate as a low-speed device, and to use the internal 
; pull-up resistor.
; 
; ============================================================================
;
; Software Licence Agreement:
; 
; THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO WARRANTIES, WHETHER 
; EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED 
; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY 
; TO THIS SOFTWARE. THE AUTHOR SHALL NOT, UNDER ANY CIRCUMSTANCES, BE LIABLE 
; FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
; 
#include <p18F4550.inc>

#include "ENGR2210.inc"

;			__CONFIG	_CONFIG1L, _PLLDIV_1_1L & _CPUDIV_OSC3_PLL4_1L & _USBDIV_2_1L
;			__CONFIG	_CONFIG1H, _FOSC_XTPLL_XT_1H & _FCMEM_OFF_1H & _IESO_OFF_1H
;			__CONFIG	_CONFIG2L, _PWRT_OFF_2L & _BOR_ON_2L & _BORV_21_2L & _VREGEN_ON_2L
;			__CONFIG	_CONFIG2H, _WDT_OFF_2H & _WDTPS_32768_2H
;			__CONFIG	_CONFIG3H, _MCLRE_ON_3H & _LPT1OSC_OFF_3H & _PBADEN_OFF_3H & _CCP2MX_ON_3H
;			__CONFIG	_CONFIG4L, _STVREN_ON_4L & _LVP_OFF_4L & _ICPRT_OFF_4L & _XINST_OFF_4L & _DEBUG_ON_4L
;			__CONFIG	_CONFIG5L, _CP0_OFF_5L & _CP1_OFF_5L & _CP2_OFF_5L & _CP3_OFF_5L
;			__CONFIG	_CONFIG5H, _CPB_OFF_5H & _CPD_OFF_5H
;			__CONFIG	_CONFIG6L, _WRT0_OFF_6L & _WRT1_OFF_6L & _WRT2_OFF_6L & _WRT3_OFF_6L
;			__CONFIG	_CONFIG6H, _WRTB_OFF_6H & _WRTC_OFF_6H & _WRTD_OFF_6H
;			__CONFIG	_CONFIG7L, _EBTR0_OFF_7L & _EBTR1_OFF_7L & _EBTR2_OFF_7L & _EBTR3_OFF_7L
;			__CONFIG	_CONFIG7H, _EBTRB_OFF_7H

#define	SHOW_ENUM_STATUS

#define	EP0_SIZE	8	; don't mess with this - may be hardcoded some places...
#define BULK_SIZE	8	; best not to mess with this either, for now


#include "usb_defs.inc"

bank0		udata
USB_buffer_desc	res		4
USB_buffer_data	res		8
USB_error_flags	res 	1
USB_curr_config	res		1
USB_device_status	res	1
USB_protocol	res		1
USB_idle_rate	res		1
USB_dev_req	res			1
USB_address_pending	res	1
USB_desc_ptr	res		1
USB_bytes_left	res		1
USB_loop_index	res		1
USB_packet_length	res	1
USB_USTAT	res			1
USB_USWSTAT	res			1
COUNTER		res			1
LED_states	res			1
Key_buffer	res			8

		; used in hex conversion
HEXTEMP		res			1
HEXTEMP1	res			1

		; mine
SCRATCH0	res			1


BOOT		code	0x0000
			goto		Main					; Reset vector
			nop
			nop
			goto		$						; High-priority interrupt vector trap
			nop
			nop
			nop
			nop
			nop
			nop
			goto		$						; Low-priority interrupt vector trap




STARTUP		code		0x0200
			goto		Main					; Reset vector
			nop
			nop
int:
			goto		int;$						; High-priority interrupt vector trap
			nop
			nop
			nop
			nop
			nop
			nop
			goto		int;$						; Low-priority interrupt vector trap


USBSTUFF	code
Descriptor
			movlw		upper Descriptor_begin
			movwf		TBLPTRU, ACCESS
			movlw		high Descriptor_begin
			movwf		TBLPTRH, ACCESS
			movlw		low Descriptor_begin
			banksel		USB_desc_ptr
			addwf		USB_desc_ptr, W, BANKED
			ifset STATUS, C, ACCESS
				incf		TBLPTRH, F, ACCESS
				ifset STATUS, Z, ACCESS
					incf		TBLPTRU, F, ACCESS
				endi
			endi
			movwf		TBLPTRL, ACCESS
			tblrd*
			movf		TABLAT, W
			return

Descriptor_begin
Device
			db			0x12, DEVICE		; bLength, bDescriptorType
			db			0x10, 0x01			; bcdUSB (low byte), bcdUSB (high byte)
			db			0x00, 0x00			; bDeviceClass, bDeviceSubClass
			db			0x00, EP0_SIZE		; bDeviceProtocl, bMaxPacketSize
			db			0x49, 0x0B			; idVendor (low byte), idVendor (high byte)
			db			0x4F, 0x06			; idProduct (low byte), idProduct (high byte)
			db			0x01, 0x00			; bcdDevice (low byte), bcdDevice (high byte)
			db			0x01, 0x02			; iManufacturer, iProduct
			db			0x00, NUM_CONFIGURATIONS	; iSerialNumber (none), bNumConfigurations

Configuration1
			db			0x09, CONFIGURATION	; bLength, bDescriptorType
			db			0x20, 0x00			; wTotalLength (low byte), wTotalLength (high byte)
			db			NUM_INTERFACES, 0x01	; bNumInterfaces, bConfigurationValue
			db			0x00, 0x80			; iConfiguration (none), bmAttributes
			db			0xFA, 0x09			; bMaxPower (100 mA), bLength (Interface1 descriptor starts here)
			db			INTERFACE, 0x00		; bDescriptorType, bInterfaceNumber
			db			0x00, 0x01			; bAlternateSetting, bNumEndpoints (excluding EP0)
			db			0x00, 0x00			; bInterfaceClass (255 = vendor-specific), bInterfaceSubClass (also vendor?)
			db			0x00, 0x00			; bInterfaceProtocol (Keyboard protocol), iInterface (none)

Interface_EP1
			db			0x07, 0x05			; bLength, bDescriptorType (0x05 = Endpoint)
			db			0x81, 0x03			; bEndpointAddress (0x81 = 01 In), bmAttributes (xxxxxx10 = Bulk)
			db			BULK_SIZE, 0x00			; wMaxPacketSize (8 bytes)
			db			0x0A, 0x07			; bInterval (don't-care for bulk & control EPs), bLength (start of EP2 descriptor!)
;Interface_EP2
;			db			0x05, 0x01
;			db			0x02, BULK_SIZE
;			db			0x00, 0xff




;;;;;;
HID1
Report1


String0
			db			String1-String0, STRING	; bLength, bDescriptorType
			db			0x09, 0x04			; wLANGID[0] (low byte), wLANGID[0] (high byte)
String1
			db			String2-String1, STRING	; bLength, bDescriptorType
			db			'D', 0x00			; bString
			db			'r', 0x00
			db			'm', 0x00
			db			'n', 0x00
			db			'4', 0x00
			db			'e', 0x00
			db			'a', 0x00
			db			' ', 0x00
			db			'T', 0x00
			db			'e', 0x00
			db			'c', 0x00
			db			'h', 0x00
			db			'.', 0x00
String2
			db			Descriptor_end-String2, STRING	; bLength, bDescriptorType
			db			'D', 0x00			; bString
			db			'r', 0x00
			db			'm', 0x00
			db			'n', 0x00
			db			' ', 0x00
			db			'T', 0x00
			db			'r', 0x00
			db			'a', 0x00
			db			'n', 0x00
			db			'c', 0x00
			db			'e', 0x00
			db			' ', 0x00
			db			'V', 0x00
			db			'i', 0x00
			db			'b', 0x00
			db			'e', 0x00
Descriptor_end


STRINGMACRO	code
#include "string_m.inc"


InitUSB
			clrf		UIE, ACCESS				; mask all USB interrupts
			clrf		UIR, ACCESS				; clear all USB interrupt flags
			movlw		0x10
			movwf		UCFG, ACCESS			; configure USB for low-speed transfers and to use the on-chip transciever and pull-up resistor
			movlw		0x08
			movwf		UCON, ACCESS			; enable the USB module and its supporting circuitry
			banksel		USB_curr_config
			clrf		USB_curr_config, BANKED
			clrf		USB_idle_rate, BANKED
			clrf		USB_USWSTAT, BANKED		; default to powered state
			movlw		0x01
			movwf		USB_device_status, BANKED
			movwf		USB_protocol, BANKED	; default protocol to report protocol initially
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED		; No device requests in process
#ifdef SHOW_ENUM_STATUS
			clrf		TRISB, ACCESS			; set all bits of PORTB as outputs
			movlw		0x01
			movwf		PORTB, ACCESS			; set bit zero to indicate Powered status
#endif
			return

ServiceUSB
			select
				caseset	UIR, UERRIF, ACCESS
					clrf		UEIR, ACCESS
					break
				caseset UIR, SOFIF, ACCESS
					bcf			UIR, SOFIF, ACCESS
					break
				caseset	UIR, IDLEIF, ACCESS
					bcf			UIR, IDLEIF, ACCESS
					bsf			UCON, SUSPND, ACCESS
					sendstring	str_suspend
#ifdef SHOW_ENUM_STATUS
					movlw		0xE0
					andwf		PORTB, F, ACCESS
					bsf			PORTB, 4, ACCESS
#endif
					break
				caseset UIR, ACTVIF, ACCESS
					bcf			UIR, ACTVIF, ACCESS
					bcf			UCON, SUSPND, ACCESS
#ifdef SHOW_ENUM_STATUS
					movlw		0xE0
					andwf		PORTB, F, ACCESS
					banksel		USB_USWSTAT
					movf		USB_USWSTAT, W, BANKED
					select
						case POWERED_STATE
							movlw	0x01
							break
						case DEFAULT_STATE
							movlw	0x02
							break
						case ADDRESS_STATE
							movlw	0x04
							break
						case CONFIG_STATE
							movlw	0x08
					ends
					iorwf		PORTB, F, ACCESS
#endif
					break
				caseset	UIR, STALLIF, ACCESS
					bcf			UIR, STALLIF, ACCESS
					break
				caseset	UIR, URSTIF, ACCESS
					banksel		USB_curr_config
					clrf		USB_curr_config, BANKED
					bcf			UIR, TRNIF, ACCESS		; clear TRNIF four times to clear out the USTAT FIFO
					bcf 		UIR, TRNIF, ACCESS
					bcf			UIR, TRNIF, ACCESS
					bcf			UIR, TRNIF, ACCESS
					clrf		UEP0, ACCESS			; clear all EP control registers to disable all endpoints
					clrf		UEP1, ACCESS
					clrf		UEP2, ACCESS
					clrf		UEP3, ACCESS
					clrf		UEP4, ACCESS
					clrf		UEP5, ACCESS
					clrf		UEP6, ACCESS
					clrf		UEP7, ACCESS
					clrf		UEP8, ACCESS
					clrf		UEP9, ACCESS
					clrf		UEP10, ACCESS
					clrf		UEP11, ACCESS
					clrf		UEP12, ACCESS
					clrf		UEP13, ACCESS
					clrf		UEP14, ACCESS
					clrf		UEP15, ACCESS

					; Begin EP setup...
					banksel		BD0OBC

					movlw		EP0_SIZE				; 
					movwf		BD0OBC, BANKED			; BD0 byte count
					movlw		low EP0_BUFF_OUT		; EP0 OUT gets a buffer...
					movwf		BD0OAL, BANKED
					movlw		high EP0_BUFF_OUT
					movwf		BD0OAH, BANKED			; ...set up its address
					movlw		0x88					; set UOWN bit (USB can write)
					movwf		BD0OST, BANKED
					movlw		low (EP0_BUFF_IN)	; EP0 IN gets a buffer...
					movwf		BD0IAL, BANKED
					movlw		high (EP0_BUFF_IN)
					movwf		BD0IAH, BANKED			; ...set up its address
					movlw		0x08					; clear UOWN bit (MCU can write)
					movwf		BD0IST, BANKED

						; setup EP1 In/Out buffers...
					movlw		BULK_SIZE				; 
					movwf		BD1OBC, BANKED			; BD1 byte count
					movlw		low EP1_BUFF_OUT		; EP1 OUT gets a buffer...
					movwf		BD1OAL, BANKED
					movlw		high EP1_BUFF_OUT
					movwf		BD1OAH, BANKED			; ...set up its address
					movlw		0x88					; set UOWN bit (USB can write)
					movwf		BD1OST, BANKED
					movlw		low (EP1_BUFF_IN)		; EP1 IN gets a buffer...
					movwf		BD1IAL, BANKED
					movlw		high (EP1_BUFF_IN)
					movwf		BD1IAH, BANKED			; ...set up its address
					movlw		0x08					; clear UOWN bit (MCU can write)
					movwf		BD1IST, BANKED


					clrf		UADDR, ACCESS			; set USB Address to 0
					clrf		UIR, ACCESS				; clear all the USB interrupt flags
					movlw		ENDPT_CONTROL
					movwf		UEP0, ACCESS			; EP0 is a control pipe and requires an ACK
					movlw		0xFF					; enable all error interrupts
					movwf		UEIE, ACCESS
					banksel		USB_USWSTAT
					movlw		DEFAULT_STATE
					movwf		USB_USWSTAT, BANKED
					movlw		0x01
					movwf		USB_device_status, BANKED	; self powered, remote wakeup disabled
#ifdef SHOW_ENUM_STATUS
					movlw		0xE0
					andwf		PORTB, F, ACCESS
					bsf 		PORTB, 1, ACCESS		; set bit 1 of PORTB to indicate Powered state
#endif
					sendstring	str_reset
					break
				caseset	UIR, TRNIF, ACCESS
					movlw		0x04
					movwf		FSR0H, ACCESS
					movf		USTAT, W, ACCESS
					andlw		0x7C					; mask out bits 0, 1, and 7 of USTAT
					movwf		FSR0L, ACCESS
					banksel		USB_buffer_desc
					movf		POSTINC0, W
					movwf		USB_buffer_desc, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+1, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+2, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+3, BANKED
					movf		USTAT, W, ACCESS
					movwf		USB_USTAT, BANKED		; save the USB status register
					bcf			UIR, TRNIF, ACCESS		; clear TRNIF interrupt flag
#ifdef SHOW_ENUM_STATUS
					andlw		0x18					; extract EP bits
					select
						case EP0
							movlw		0x20
							break
						case EP1
							movlw		0x40
							break
						case EP2
							movlw		0x80
							break
					ends
					xorwf		PORTB, F, ACCESS		; toggle bit 5, 6, or 7 of PORTB to reflect EP activity
#endif
					movf		USB_buffer_desc, W, BANKED
					andlw		0x3C					; extract PID bits
					select
						case TOKEN_SETUP
							call		ProcessSetupToken
							break
						case TOKEN_IN
							call		ProcessInToken
							break
						case TOKEN_OUT
							call		ProcessOutToken
							break
					ends
					break
			ends
			return

ProcessSetupToken
			;sendstring	str_ProcessSetupToken

			banksel		USB_buffer_data
			movf		USB_buffer_desc+ADDRESSH, W, BANKED
			movwf		FSR0H, ACCESS
			movf		USB_buffer_desc+ADDRESSL, W, BANKED
			movwf		FSR0L, ACCESS
			movf		POSTINC0, W
			movwf		USB_buffer_data, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+1, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+2, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+3, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+4, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+5, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+6, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+7, BANKED
			banksel		BD0OBC
			movlw		0x08
			movwf		BD0OBC, BANKED					; reset the byte count
			movwf		BD0IST, BANKED					; return the in buffer to us (dequeue any pending requests)
			banksel		USB_buffer_data+bmRequestType
			ifl	USB_buffer_data+bmRequestType, ==, 0x21	; host to device
				movlw		0xC8
			otherwise				; device to host
				movlw		0x88
			endi
			banksel		BD0OST
			movwf		BD0OST, BANKED					; set EP0 OUT UOWN back to USB and DATA0/DATA1 packet according to request type
			bcf			UCON, PKTDIS, ACCESS			; assuming there is nothing to dequeue, clear the packet disable bit
			banksel		USB_dev_req
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED				; clear the device request in process
			movf		USB_buffer_data+bmRequestType, W, BANKED
			andlw		0x60							; extract request type bits
			select
				case STANDARD
					sendstring str_std_req
					call		StandardRequests
					break
				case CLASS
					sendstring	str_class_req
					call		ClassRequests
					break
				case VENDOR
					call		VendorRequests
					break
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			ends
			return

StandardRequests
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_STATUS
					sendstring	str_get_status
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							sendstring	str_rc_device
							banksel		BD0IAH
							movf		BD0IAH, W, BANKED
							movwf		FSR0H, ACCESS
							movf		BD0IAL, W, BANKED				; get buffer pointer
							movwf		FSR0L, ACCESS
							banksel		USB_device_status
							movf		USB_device_status, W, BANKED	; copy device status byte to EP0 buffer
							movwf		POSTINC0
							clrf		INDF0
							banksel		BD0IBC
							movlw		0x02
							movwf		BD0IBC, BANKED					; set byte count to 2
							movlw		0xC8
							movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
							break
						case RECIPIENT_INTERFACE
							sendstring	str_rc_interface
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
									break
								case CONFIG_STATE
									ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED				; get buffer pointer
										movwf		FSR0L, ACCESS
										clrf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit																								
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
							ends
							break
						case RECIPIENT_ENDPOINT
							sendstring	str_rc_ep
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED				; put EP0 IN buffer pointer...
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED
										movwf		FSR0L, ACCESS					; ...into FSR0
										ifset UEP0, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
								case CONFIG_STATE
									banksel		BD0IAH
									movf		BD0IAH, W, BANKED					; put EP0 IN buffer pointer...
									movwf		FSR0H, ACCESS
									movf		BD0IAL, W, BANKED
									movwf		FSR0L, ACCESS						; ...into FSR0
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR1H, ACCESS
									movlw		low UEP0
									movwf		FSR1L, ACCESS						; ...into FSR1
									banksel		USB_buffer_data+wIndex
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP and...
									andlw		0x0F								; ...strip off direction bit
									ifclr PLUSW1, EPOUTEN, ACCESS
									andifclr PLUSW1, EPINEN, ACCESS
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									otherwise
										ifset PLUSW1, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										banksel		BD0IBC
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							ends
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case CLEAR_FEATURE
				case SET_FEATURE
					sendstring	str_feature
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							sendstring	str_rc_device
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case DEVICE_REMOTE_WAKEUP
									ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
										bcf			USB_device_status, 1, BANKED
									otherwise
										bsf			USB_device_status, 1, BANKED
									endi
									banksel		BD0IBC
									clrf		BD0IBC, BANKED					; set byte count to 0
									movlw		0xC8
									movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									break
								default
									bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
							ends
							break
						case RECIPIENT_ENDPOINT
							sendstring	str_rc_ep
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
											bcf			UEP0, EPSTALL, ACCESS
										otherwise
											bsf			UEP0, EPSTALL, ACCESS
										endi
										banksel		BD0IBC
										clrf		BD0IBC, BANKED					; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
								case CONFIG_STATE
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR0H, ACCESS
									movlw		low UEP0
									movwf		FSR0L, ACCESS						; ...into FSR0
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									addwf		FSR0L, F, ACCESS					; add EP number to FSR0
									ifset		STATUS, C, ACCESS
										incf		FSR0H, F, ACCESS
									endi
									ifclr INDF0, EPOUTEN, ACCESS
									andifclr INDF0, EPINEN, ACCESS
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									otherwise
										ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
											bcf			INDF0, EPSTALL, ACCESS
										otherwise
											bsf			INDF0, EPSTALL, ACCESS
										endi
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							ends
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_ADDRESS
						sendstring	str_set_address
					ifset USB_buffer_data+wValue, 7, BANKED		; if new device address is illegal, send Request Error
						bsf			UEP0, EPSTALL, ACCESS		; set EP0 protocol stall bit to signify Request Error
					otherwise
						movlw		SET_ADDRESS
						movwf		USB_dev_req, BANKED			; processing a SET_ADDRESS request
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_address_pending, BANKED	; save new address
						banksel		BD0IBC
						clrf		BD0IBC, BANKED				; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED				; send packet as DATA1, set UOWN bit
					endi
					break
				case GET_DESCRIPTOR
					movwf		USB_dev_req, BANKED				; processing a GET_DESCRIPTOR request
					sendstring	str_get_desc
					movf		USB_buffer_data+(wValue+1), W, BANKED
					select
						case DEVICE
							sendstring	str_desc_dev
							movlw		low (Device-Descriptor_begin)
							movwf		USB_desc_ptr, BANKED
							call		Descriptor				; get descriptor length
							movwf		USB_bytes_left, BANKED
							ifl USB_buffer_data+(wLength+1), ==, 0
							andiff USB_buffer_data+wLength, <, USB_bytes_left
								movf		USB_buffer_data+wLength, W, BANKED
								movwf		USB_bytes_left, BANKED
							endi
							call		SendDescriptorPacket
							break
						case CONFIGURATION
							sendstring	str_desc_cfg
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (Configuration1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								addlw		0x02				; add offset for wTotalLength
								movwf		USB_desc_ptr, BANKED
								call		Descriptor			; get total descriptor length
								movwf		USB_bytes_left, BANKED
								movlw		0x02
								subwf		USB_desc_ptr, F, BANKED	; subtract offset for wTotalLength
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case STRING
							sendstring	str_desc_string
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (String0-Descriptor_begin)
									break
								case 1
									movlw		low (String1-Descriptor_begin)
									break
								case 2
									movlw		low (String2-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor		; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case HID
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (HID1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor				; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case REPORT
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		0x3F
									movwf		USB_bytes_left, BANKED	; set descriptor length
									movlw		low (Report1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case GET_CONFIGURATION
					sendstring	str_get_cfg
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS
					banksel		USB_curr_config
					movf		USB_curr_config, W, BANKED
					movwf		INDF0					; copy current device configuration to EP0 IN buffer
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set EP0 IN byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_CONFIGURATION
					sendstring	str_set_cfg
					ifl USB_buffer_data+wValue, <=, NUM_CONFIGURATIONS
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_curr_config, BANKED
						select
							case 0
								movlw		ADDRESS_STATE
								movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
								movlw		0xE0
								andwf		PORTB, F, ACCESS
								bsf			PORTB, 2, ACCESS
#endif
								break
							default
								movlw		CONFIG_STATE
								movwf		USB_USWSTAT, BANKED
								movlw		0x08

								banksel		BD1IBC
									; config EP1 IN
								movwf		BD1IBC, BANKED			; set EP1 IN byte count to 8 
								movlw		low EP1_BUFF_IN ;(USB_Buffer+0x10)
								movwf		BD1IAL, BANKED			; set EP1 IN buffer address
								movlw		high EP1_BUFF_IN ;(USB_Buffer+0x10)
								movwf		BD1IAH, BANKED
								movlw		0x48
								movwf		BD1IST, BANKED			; clear UOWN bit (PIC can write EP1 IN buffer)
								;movlw		ENDPT_NON_CONTROL
								;movwf		UEP1, ACCESS			; enable EP1 for in/out xfer
									; redundant, set at end of this section...

									; config EP1 OUT
								movlw		0x08
								movwf		BD1OBC, BANKED			; set EP1 OUT byte count to 8 
								movlw		low EP1_BUFF_OUT ;(USB_Buffer+0x10)
								movwf		BD1OAL, BANKED			; set EP1 IN buffer address
								movlw		high EP1_BUFF_OUT ;(USB_Buffer+0x10)
								movwf		BD1OAH, BANKED
								movlw		0x88
								movwf		BD1OST, BANKED			; set UOWN bit (SIE can write EP1 OUT buffer)
								movlw		ENDPT_NON_CONTROL
								movwf		UEP1, ACCESS			; enable EP1 for in/out xfer


#ifdef SHOW_ENUM_STATUS
								movlw		0xE0
								andwf		PORTB, F, ACCESS
								bsf			PORTB, 3, ACCESS
#endif
						ends
						banksel		BD0IBC
						clrf		BD0IBC, BANKED			; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					otherwise
						bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					endi
					break
				case GET_INTERFACE
					sendstring	str_get_interface
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
								banksel		BD0IAH
								movf		BD0IAH, W, BANKED
								movwf		FSR0H, ACCESS
								movf		BD0IAL, W, BANKED		; get buffer pointer
								movwf		FSR0L, ACCESS
								clrf		INDF0					; always send back 0 for bAlternateSetting
								movlw		0x01
								movwf		BD0IBC, BANKED			; set byte count to 1
								movlw		0xC8
								movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_INTERFACE
					sendstring	str_set_interface
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
								movf		USB_buffer_data+wValue, W, BANKED
								select
									case 0									; currently support only bAlternateSetting of 0
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
										break
									default
										bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
								ends
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_DESCRIPTOR
				case SYNCH_FRAME
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					break
			ends
			return

ClassRequests
			sendstring	str_class_req
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_REPORT
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		Key_buffer
					movf		Key_buffer, W, BANKED	; copy modifier byte...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+1, W, BANKED	; copy reserved byte...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+2, W, BANKED	; copy keycode 1...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+3, W, BANKED	; copy keycode 2...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+4, W, BANKED	; copy keycode 3...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+5, W, BANKED	; copy keycode 4...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+6, W, BANKED	; copy keycode 5...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		Key_buffer+7, W, BANKED	; copy keycode 6...
					movwf		INDF0					; ...to EP0 IN buffer
					banksel		BD0IBC
					movlw		0x08
					movwf		BD0IBC, BANKED			; set EP0 IN buffer byte count to 8
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_REPORT
					movwf		USB_dev_req, BANKED		; processing a SET_REPORT request
					break
				case GET_PROTOCOL
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		USB_protocol
					movf		USB_protocol, W, BANKED
					movwf		INDF0
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_PROTOCOL
					movf		USB_buffer_data+wValue, W, BANKED
					movwf		USB_protocol, BANKED	; update the new protocol value
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case GET_IDLE
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		USB_idle_rate
					movf		USB_idle_rate, W, BANKED
					movwf		INDF0
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_IDLE
					movf		USB_buffer_data+wValue, W, BANKED
					movwf		USB_idle_rate, BANKED	; update the new idle rate
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			ends
			return

VendorRequests
			;sendstring	str_vendor_req

			; HACK: FIXME: Just testing if I can read/write without Windows flagging an error
			movf		USB_buffer_data+0, W, BANKED	; bmRequestType
			btfss		WREG, 7							; data transfer direction
			goto		vendor_h2d						; if "Out" (host -> device)

		vendor_d2h:
			;sendstring	str_d2h
			movf		USB_buffer_data+bRequest, W, BANKED
			call		SENDHEX							; show request to serial

			banksel		BD0IAH
			movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD0IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			;banksel		USB_idle_rate
			;movf		USB_idle_rate, W, BANKED
			;movwf		INDF0

			;movf		USB_buffer_data+bRequest, W, BANKED	; test: spit request back to PC
			;movwf		INDF0					; into Out buffer

			call		setbuf					; set EP0in contents

			;banksel		BD0IBC
			;movlw		0x07
			;movwf		BD0IBC, BANKED			; set byte count to 8
			;movlw		0xC8
			;movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit

			banksel		BD0IBC
			movlw		0x07
			movwf		BD0IBC, BANKED			; set EP1 IN buffer byte count to 8
			movf		BD0IST, W, BANKED		; get EP1 IN status register
			andlw		0x40					; extract the DATA01 bit
			xorlw		0x40					; toggle the DATA01 bit
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD0IST, BANKED			; send packet



			return

		vendor_h2d:
			;sendstring	str_h2d
			;select
			;	default
			;		bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			;ends

			; HACK: FIXME: Just testing if I can read/write without Windows flagging an error
;			movf		USB_buffer_data+bRequest, W, BANKED
;			call		SENDHEX							; show request to serial
;			xorlw	H'01'	; request 0x01: set vibe
;			bnz		set_vibe_rate_done	; if no

;			sendstring	str_set_vibe_rate
			movf		USB_buffer_data+wValue, W, BANKED
			movwf		PORTA	; debug lites
			movwf		PORTB	; debug lites
			movwf		CCPR1L	; set PWM duty cycle
;			call		SENDHEX
;			call		SENDCMNDP2CR



		set_vibe_rate_done:
			call		show_EP0OUT		; debug

			banksel		BD0IBC
			clrf		BD0IBC, BANKED			; set byte count to 0
			movlw		0xC8
			movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit

			return

ProcessInToken
			;sendstring	str_process_in
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_ADDRESS
							sendstring	str_addressed
							movf		USB_address_pending, W, BANKED
							movwf		UADDR, ACCESS
							select
								case 0
									movlw		DEFAULT_STATE
									movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
									movlw		0xE0
									andwf		PORTB, F, ACCESS
									bsf			PORTB, 1, ACCESS
#endif
									break
								default
									movlw		ADDRESS_STATE
									movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
									movlw		0xE0
									andwf		PORTB, F, ACCESS
									bsf			PORTB, 2, ACCESS
#endif
							ends
							break
						case GET_DESCRIPTOR
							sendstring	str_tkn_get_desc
							call		SendDescriptorPacket
							break
					ends
					break
				case EP1
						sendstring	str_ep1_in
						; FIXME: BULK In handler here
					break
				case EP2
						; WTF is EP2?
					break
			ends
			return

ProcessOutToken
			sendstring	str_process_out
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_REPORT
							movlw		NO_REQUEST
							movwf		USB_dev_req, BANKED	; clear device request
							banksel		BD0OAH
							movf		BD0OAH, W, BANKED	; put EP0 OUT buffer pointer...
							movwf		FSR0H, ACCESS
							movf		BD0OAL, W, BANKED
							movwf		FSR0L, ACCESS		; ...into FSR0
							movf		INDF0, W			; get the first byte in the buffer and...
							banksel		LED_states
							movwf		LED_states, BANKED	; ...update the LED states with it
							ifset LED_states, 1, BANKED		; set RA1 to reflect the new state of the CAPS LOCK LED
								bsf			PORTA, 1, ACCESS
							otherwise
								bcf			PORTA, 1, ACCESS
							endi
					ends

					movlw		NO_REQUEST
					movwf		USB_dev_req, BANKED	; clear device request ; FIXME: CHECKME: what's this?
					banksel		BD0OAH
					movf		BD0OAH, W, BANKED	; put EP0 OUT buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0OAL, W, BANKED
					movwf		FSR0L, ACCESS		; ...into FSR0

					;call		show_EP0OUT			; debug
					call		showbuf				; debug

					banksel		BD0OBC
					movlw		0x08
					movwf		BD0OBC, BANKED
					movlw		0x88
					movwf		BD0OST, BANKED
					clrf		BD0IBC, BANKED		; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED		; send packet as DATA1, set UOWN bit
					break
				case EP1
						sendstring	str_ep1_out
						; debug stuff...
						lfsr	FSR0, EP1_BUFF_OUT	; (Host to Device)
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						movf	POSTINC0, w
						call	SENDCMNDP2
						call	SENDCMNDP2CR
						; FIXME: BULK Out handler here

						banksel		BD1OBC
						movlw		0x08
						movwf		BD1OBC, BANKED
						movlw		0x88
						movwf		BD1OST, BANKED
						clrf		BD1IBC, BANKED		; set byte count to 0
						movlw		0xC8
						movwf		BD1IST, BANKED		; send packet as DATA1, set UOWN bit


					break
				case EP2
					break
			ends
			return

SendDescriptorPacket
			banksel		USB_bytes_left
			ifl USB_bytes_left, <, 8
				movlw		NO_REQUEST
				movwf		USB_dev_req, BANKED		; sending a short packet, so clear device request
				movf		USB_bytes_left, W, BANKED
			otherwise
				movlw		0x08
			endi
			subwf		USB_bytes_left, F, BANKED
			movwf		USB_packet_length, BANKED
			banksel		BD0IBC
			movwf		BD0IBC, BANKED			; set EP0 IN byte count with packet size
			movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD0IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		USB_loop_index
			forlf USB_loop_index, 1, USB_packet_length
				call		Descriptor			; get next byte of descriptor being sent
				movwf		POSTINC0			; copy to EP0 IN buffer, and increment FSR0
				incf		USB_desc_ptr, F, BANKED	; increment the descriptor pointer
			next USB_loop_index
			banksel		BD0IST
			movlw		0x40
			xorwf		BD0IST, W, BANKED		; toggle the DATA01 bit
			andlw		0x40					; clear the PIDs bits
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD0IST, BANKED
			return

SendKeyBuffer
			banksel		BD1IAH
			movf		BD1IAH, W, BANKED		; put EP1 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD1IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		Key_buffer
			movf		Key_buffer, W, BANKED	; copy modifier byte...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+1, W, BANKED	; copy reserved byte...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+2, W, BANKED	; copy keycode 1...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+3, W, BANKED	; copy keycode 2...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+4, W, BANKED	; copy keycode 3...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+5, W, BANKED	; copy keycode 4...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+6, W, BANKED	; copy keycode 5...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		Key_buffer+7, W, BANKED	; copy keycode 6...
			movwf		INDF0					; ...to EP1 IN buffer
			banksel		BD1IBC
			movlw		0x08
			movwf		BD1IBC, BANKED			; set EP1 IN buffer byte count to 8
			movf		BD1IST, W, BANKED		; get EP1 IN status register
			andlw		0x40					; extract the DATA01 bit
			xorlw		0x40					; toggle the DATA01 bit
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD1IST, BANKED			; send packet
			return



APPLICATION	code
Main
			movlw	B'00000000'	; external
			movwf	OSCCON

			clrf	WDTCON		; kill watchdog timer

			call	epwm_init

			call	INITUSART
			banksel		COUNTER
			for COUNTER, 0x01, 0x1E				; do nothing for a little while
			next COUNTER
			clrf		PORTA, ACCESS
			movlw		0x0F
			movwf		ADCON1, ACCESS			; set up PORTA to be digital I/Os
			clrf		TRISA, ACCESS			; set up all PORTA pins to be digital outputs
			movlw		0x01
			movwf		TRISA, ACCESS			; set up RA0 as an input and RA1-RA5 as outputs
			movlw		0xC7
			movwf		T0CON, ACCESS			; set prescaler for Timer0 for 1:256 scaling (Timer0 will go off every 10.9 ms)
test1:
			sendstring	str_hello		; test USART
;			goto	test1

			;call	GETCMD

			call		InitUSB					; initialize the USB registers and serial interface engine
			repeat
				call		ServiceUSB			; service USB requests...
				banksel		USB_USWSTAT
			until USB_USWSTAT, ==, CONFIG_STATE	; ...until the host configures the peripheral
			banksel		COUNTER
			clrf		COUNTER, BANKED
			clrf		LED_states, BANKED
			clrf		Key_buffer, BANKED
			clrf		Key_buffer+1, BANKED
			clrf		Key_buffer+2, BANKED
			clrf		Key_buffer+3, BANKED
			clrf		Key_buffer+4, BANKED
			clrf		Key_buffer+5, BANKED
			clrf		Key_buffer+6, BANKED
			clrf		Key_buffer+7, BANKED
			repeat
				repeat
					call		ServiceUSB			; service USB requests...
				untilset INTCON, T0IF, ACCESS		; ...until Timer0 goes off
				bcf			INTCON, T0IF, ACCESS	; clear Timer0 interrupt flag
				banksel		BD1IST
				ifclr BD1IST, UOWN, BANKED			; check to see if the PIC owns the EP1 IN buffer
					movlw		high (Key_buffer+2)
					movwf		FSR0H, ACCESS
					movlw		low (Key_buffer+2)
					movwf		FSR0L, ACCESS		; set FSR0 to point to start of keycodes in Key_buffer
					ifset PORTA, 0, ACCESS			; see if RA0 is high (i.e., CAPS LOCK key is pressed)
						movlw		0x39			; USB keycode for CAPS LOCK
						movwf		POSTINC0		; put CAPS LOCK keycode into Key_buffer
					endi
					call		GetNextKeycode		; get the next keycode and...
					movwf		POSTINC0			; ...put it into Key_buffer
					clrf		INDF0
					incf		COUNTER, F, BANKED	; increment COUNTER...
					movlw		0x07
					andwf		COUNTER, F, BANKED	; ...modulo 8
					;call		SendKeyBuffer
				endi
			forever

GetNextKeycode
			movlw		upper KeycodeTable
			movwf		TBLPTRU, ACCESS
			movlw		high KeycodeTable
			movwf		TBLPTRH, ACCESS
			movlw		low KeycodeTable
			banksel		COUNTER
			addwf		COUNTER, W, BANKED
			ifset STATUS, C, ACCESS
				incf		TBLPTRH, F, ACCESS
				ifset STATUS, Z, ACCESS
					incf		TBLPTRU, F, ACCESS
				endi
			endi
			movwf		TBLPTRL, ACCESS
			tblrd*
			movf		TABLAT, W, ACCESS
			return


#include "4220uart.inc"
#include "lcmhex01.inc"
#include "tmr2.inc"
#include "4220pwm.inc"
#include "test.inc"



KeycodeTable
			db			0x09, 0x12	; USB keycode for 'f', USB keycode for 'o'
			db			0x00, 0x12	; USB keycode for indicating no event, USB keycode for 'o'
			db			0x05, 0x04	; USB keycode for 'b', USB keycode for 'a'
			db			0x15, 0x2C	; USB keycode for 'r', USB keycode for ' '

str_hello
	dw	"\r\n\r\n\r\nHuzzah World\r\n\x00"


str_suspend:
	DW	"SUSPEND\r\n\x00"

str_reset:
	DW	"RESET\r\n\x00"

str_ProcessSetupToken:
	DW	"ProcessSetupToken\r\n\x00"

str_std_req:
	DW	"Std. Req.\r\n\x00"

;str_class_req:
;	DW	"Class Req.\r\n\x00"

str_vendor_req:
	DW	"Vendor Req.\r\n\x00"

str_set_address:
	DW	"Set Address\r\n\x00"

str_addressed:
	DW	"ADDRESSED\r\n\x00"

str_get_desc:
	DW	"GetDesc: \x00"

str_desc_dev:
	DW	"Dev\r\n\x00"

str_process_in:
	DW	"ProcessInToken\r\n\x00"

str_tkn_get_desc:
	DW	"PIT:SendDescPacket\r\n\x00"

str_process_out:
	DW	"ProcessOutToken\r\n\x00"



str_get_status:
	DW	"GetStatus \x00"


str_rc_device:
	DW	"(Device)\r\n\x00"

str_rc_interface:
	DW	"(Interface)\r\n\x00"

str_rc_ep:
	DW	"(EP)\r\n\x00"

str_feature:
	DW	"Set/Clear Feature \x00"

str_desc_cfg:
	DW	"Config\r\n\x00"

str_desc_string:
	DW	"String\r\n\x00"

str_get_cfg:
	DW	"Get Config\r\n\x00"

str_set_cfg:
	DW	"Set Config\r\n\x00"

str_get_interface:
	DW	"Get Interface\r\n\x00"

str_set_interface:
	DW	"Set Interface\r\n\x00"

str_class_req:
	DW	"(Class req)\r\n\x00"

str_ep1_in:
	DW	"EP1 in request...\r\n\x00"

str_ep1_out:
	DW	"EP1 out request...\r\n\x00"

str_d2h:
	DW	"D->H\r\n\x00"

str_h2d:
	DW	"H->D\r\n\x00"



str_set_vibe_rate:
	DW	"Set vibe: \x00"

			end
